import numpy as np
from hypothesis import given, seed
from hypothesis import strategies as st

from pyrho.pgrid import PGrid


@seed(1337)
@given(
    nx=st.integers(min_value=40, max_value=80),
    ny=st.integers(min_value=40, max_value=80),
    A=st.integers(min_value=1, max_value=10),
    B=st.integers(min_value=1, max_value=10),
)
def test_pgrid(checker_2D, nx, ny, A, B):
    checker = checker_2D()
    XX, YY = checker.get_xy(np.eye(2), [nx, ny])
    ZZ = checker.function(XX, YY)
    # create PGrid
    pgrid = PGrid(ZZ, [[A, 0], [0, B]])
    assert pgrid.grid_shape == (nx, ny)

    # test reconstructing w/ upsampling
    transformed_data = pgrid._transform_data(
        np.eye(2), origin=(0.0, 0.0), grid_out=[nx, ny], up_sample=4
    )
    assert np.max(np.abs(ZZ - transformed_data)) < 1e-4

    # test reconstructing w/o upsampling
    transformed_data = pgrid._transform_data(
        np.eye(2), origin=[0, 0], grid_out=[nx, ny]
    )
    assert np.max(transformed_data) > 0.1
    assert np.max(np.abs(ZZ - transformed_data)) < 1e-4

    transformed_data = np.real(
        pgrid._transform_data(
            [[1, 1], [1, -1]], origin=[0.5, 0.5], grid_out=[54, 48], up_sample=8
        )
    )
    XX_ref, YY_ref = checker.get_xy([[1, 1], [1, -1]], [54, 48], origin=[0.5, 0.5])
    ZZ_ref = checker.function(XX_ref, YY_ref)
    assert np.max(np.abs(ZZ_ref - transformed_data)) < 1e-4

    # check transformed object
    transformed_obj = pgrid.get_transformed(
        sc_mat=[[1, 1], [1, -1]], grid_out=[54, 48], origin=[0.5, 0.5]
    )
    assert transformed_obj.grid_shape == (54, 48)



"""Shared fixtures for tests."""

from dataclasses import dataclass

import numpy as np
import pytest
from matplotlib import pyplot as plt


@pytest.fixture(scope="session")
def test_dir():
    from pathlib import Path

    module_dir = Path(__file__).resolve().parent.parent
    test_dir = module_dir / "test_files"
    return test_dir.resolve()


@dataclass
class Checker2D:
    """Check 2D regridding using an analytic 2D function."""

    mX: int = 1
    mY: int = 1

    def function(self, x, y):
        """Analytical function in 2D."""
        return np.sin(self.mX * x * 2 * np.pi)  # + np.cos(self.mY * y * 2 * np.pi)

    def get_xy(self, lat_mat, grids, origin=(0, 0)):
        """Get the x and y coordinates for a given pair of lattice vectors and grid size.

        Args:
        ----
            lat_mat: lattice vectors
            grids: grid size
            origin_cart: origin of the lattice vectors in cartesian coordinates

        Returns:
        -------
            XX: x coordinates for the grid in the shape of the grid
            YY: y coordinates for the grid in the shape of the grid

        """
        a_vec = np.linspace(0, 1, grids[0], endpoint=False)
        b_vec = np.linspace(0, 1, grids[1], endpoint=False)
        AA, BB = np.meshgrid(a_vec, b_vec, indexing="ij")
        frac = np.vstack([AA.flatten(), BB.flatten()])
        XX, YY = np.dot(np.array(lat_mat).T, frac)
        XX += origin[0]
        YY += origin[1]
        XX = XX.reshape(grids)
        YY = YY.reshape(grids)
        return XX, YY

    def check_z(self, ZZ, lat_mat, grids):
        """Check the regridded data against the analytic function."""
        XX, YY = self.get_xy(lat_mat, grids)
        ZZ_check = self.function(XX, YY)
        max_diff = np.max(ZZ - ZZ_check)
        assert max_diff < 1e-4

    def plot(self, lat_mat, grids, origin=(0, 0), ZZ=None):
        XX, YY = self.get_xy(lat_mat, grids, origin=origin)
        fig, ax = plt.subplots()
        if ZZ is None:
            ZZ = self.function(XX, YY)
        ax.pcolormesh(XX, YY, ZZ)
        ax.set_aspect("equal")
        plt.show()


@pytest.fixture(scope="session")
def checker_2D():
    return Checker2D



"""mp-pyrho package."""

__author__ = """Jimmy Shen"""
__email__ = "jmmshn@gmail.com"


from pkg_resources import DistributionNotFound, get_distribution

try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:
    pass



"""Helper functions to visualize the data in plotly."""
from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
import plotly.graph_objs as go
from matplotlib import pyplot as plt

if TYPE_CHECKING:
    from matplotlib.axes import Axes


"""Visualization functions do the scatter plots in plotly since it seems to be more efficient."""


def get_scatter_plot(
    data_in: np.ndarray,
    lat_mat: np.ndarray,
    skips: int = 1,
    logcolor: bool = False,
    mask: np.ndarray = None,
    opacity: float = 1.0,
    marker_size: int = 5,
    plotter: str = "matplotlib",
) -> go.Figure | Axes:
    """Return a plotly fig object for plotting.

    Parameters
    ----------
    data_in:
        Structured grid data to be plotted
    lat_mat:
        Lattice vectors of the cell this must be a 2d array
    skips:
        Reduction factor of the grid points for plotting, only show [::skips] in each direction
    logcolor:
        If True, assign the color in log scale
    mask:
        Filter the points to plot
    opacity:
        Opacity of each point being plotted
    marker_size:
        Size of the markers in the 3D scatter plot
    marker_size:
        Marker size for the scatter plot
    plotter:
        Plotter to use, either "matplotlib" or "plotly"


    Returns
    -------
    Figure | Axes:
        `Figure` object or `Axes` object from matplotlib to be rendered in a notebook

    """
    ndim = len(data_in.shape)
    if ndim > 3:
        raise NotImplementedError("Can only render data of 1, 2, or 3 dimensions.")

    ss = slice(0, None, skips)
    trimmed_data = np.real(data_in).copy()
    trimmed_data = trimmed_data[(ss,) * ndim]

    if mask is not None:
        flat_mask = mask[(ss,) * ndim].flatten()
    else:
        flat_mask = np.ones_like(trimmed_data, dtype=bool).flatten()

    vecs = [
        np.linspace(0, 1, trimmed_data.shape[_], endpoint=False) for _ in range(ndim)
    ]
    gridded = np.meshgrid(*vecs, indexing="ij")  # indexing to match the labeled array
    res = np.dot(lat_mat.T, [g_.flatten() for g_ in gridded])

    if logcolor:
        cc = np.log(trimmed_data.flatten())
    else:
        cc = trimmed_data.flatten()

    xx, yy, zz = None, None, None
    if ndim > 0:
        xx = res[0, flat_mask]
    if ndim > 1:
        yy = res[1, flat_mask]
    if ndim > 2:
        zz = res[2, flat_mask]

    cc = cc[flat_mask]
    if plotter == "matplotlib":
        return _scatter_matplotlib(xx, yy, zz, cc, ndim, marker_size, opacity)
    elif plotter == "plotly":
        return _scatter_plotly(xx, yy, zz, cc, ndim, marker_size, opacity)
    else:
        raise ValueError("plotter must be one of 'matplotlib' or 'plotly'")


def _scatter_plotly(xx, yy, zz, cc, ndim, marker_size, opacity):
    """Return the plotly object."""
    if ndim == 1:
        data = go.Scatter(
            x=xx,
            y=cc,
            mode="markers",
            marker=dict(
                size=marker_size,
                color="red",
            ),
        )
    if ndim == 2:
        data = go.Scatter(
            x=xx,
            y=yy,
            mode="markers",
            marker=dict(
                size=marker_size,
                color=cc,  # set color to an array/list of desired values
                colorscale="Viridis",  # choose a colorscale
                opacity=opacity,
            ),
        )
    if ndim == 3:
        data = go.Scatter3d(
            x=xx,
            y=yy,
            z=zz,
            mode="markers",
            marker=dict(
                size=marker_size,
                color=cc,
                colorscale="Viridis",
                opacity=opacity,
            ),
        )
    fig = go.Figure(data=[data])

    fig.update_layout(template="plotly_white")
    if ndim == 2:
        fig.update_layout(
            width=800, height=800, yaxis=dict(scaleanchor="x", scaleratio=1)
        )
    if ndim == 3:
        fig.update_layout(width=800, height=800, scene_aspectmode="data")
    return fig


def _scatter_matplotlib(xx, yy, zz, cc, ndim, marker_size, opacity) -> Axes:
    """Return the matplotlib object."""
    fig = plt.figure(figsize=(8, 8))
    if ndim == 1:
        ax = fig.add_subplot()
        ax.scatter(xx, cc, marker_size, color="red")
    if ndim == 2:
        ax = fig.add_subplot()
        ax.scatter(xx, yy, marker_size, c=cc)
    if ndim == 3:
        ax = fig.add_subplot(projection="3d")
        ax.scatter(xx, yy, zz, marker_size, c=cc)



"""automatically run doctests."""

import doctest

import pyrho


def test_doctests():
    """Find all modules and attempt to run doctest using pytest."""
    doctest.testmod(pyrho.utils, verbose=True, raise_on_error=True)



"""Chang Density Objects: Periodic Grid + Lattice / Atoms."""
from __future__ import annotations

import math
import warnings
from dataclasses import dataclass
from typing import TYPE_CHECKING

import numpy as np
import numpy.typing as npt
from monty.dev import deprecated
from monty.json import MSONable
from pymatgen.analysis.structure_matcher import ElementComparator, StructureMatcher
from pymatgen.core.lattice import Lattice
from pymatgen.io.vasp import Chgcar, Poscar, VolumetricData

from pyrho.pgrid import PGrid
from pyrho.utils import get_sc_interp

if TYPE_CHECKING:
    from pymatgen.core.structure import Structure


__all__ = ["ChargeDensity"]


@dataclass
class ChargeDensity(MSONable):
    """Charge density object.

    Defines a charge density with a PGrid object along with the atomic structure.

    Attributes
    ----------
    pgrids: Dict[str, PGrid]
        Dictionaries whose values are periodic ``PGrid`` objects
        representing some periodic scalar field (typically the keys are ``total`` and ``diff``
        for spin-polarized charge densities.)
    structure: Structure
        The atomic structure for the electronic charge density.
    normalization: str | None = "vasp"
        The normalzation scheme (``vasp`` or ``None``)

    """

    pgrids: dict[str, PGrid]
    structure: Structure
    normalization: str | None = "vasp"

    def __post_init__(self):
        """Post initialization.

        Steps:
            - Make sure all the lattices are identical

        """
        lattices = [self.pgrids[key].lattice for key in self.pgrids.keys()]
        if not all(
            np.allclose(self.structure.lattice.matrix, lattice) for lattice in lattices
        ):
            raise ValueError("Lattices are not identical")

    @property
    def normalized_data(self) -> dict[str, npt.NDArray]:
        """Get the normalized data.

        Since different codes use different normalization methods for
        volumetric data we should convert them to the same units (electrons / Angstrom^3)

        Returns
        -------
        dict[str, NDArray]:
            The normalized data in units of (electrons / Angstrom^3)

        """
        return {
            k: _normalize_data(
                grid_data=v.grid_data,
                lattice=self.structure.lattice,
                normalization=self.normalization,
            )
            for k, v in self.pgrids.items()
        }

    @property
    def grid_shape(self) -> tuple[int, int, int]:
        """Return the shape of the charge density."""
        return self.pgrids["total"].grid_shape

    @property
    def normalized_pgrids(self) -> dict[str, PGrid]:
        """Get the normalized pgrids.

        Since different codes use different normalization methods for
        volumetric data we should convert them to the same units (electrons / Angstrom^3)

        Returns
        -------
        dict[str, PGrid]:
            The normalized pgrids in units of (electrons / Angstrom^3)

        """
        return {
            k: PGrid(
                grid_data=_normalize_data(
                    grid_data=v.grid_data,
                    lattice=self.structure.lattice,
                    normalization=self.normalization,
                ),
                lattice=v.lattice,
            )
            for k, v in self.pgrids.items()
        }

    @property
    def lattice(self) -> npt.NDArray:
        """Lattice represented as an NDArray."""
        return self.structure.lattice.matrix

    @classmethod
    def from_pmg(
        cls, vdata: VolumetricData, normalization: str = "vasp"
    ) -> "ChargeDensity":
        """Get data from pymatgen object.

        Read a single key from the data field of a VolumetricData object

        Parameters
        ----------
        vdata:
            The volumetric data object
        normalization:
            The normalization scheme
            - 'vasp' sum of the data / number of grid points == number of electrons
            - None/"none" no normalization

        Returns
        -------
        ChargeDensity:
            The charge density object

        """
        pgrids = {
            k: PGrid(v, vdata.structure.lattice.matrix) for k, v in vdata.data.items()
        }
        return cls(
            pgrids=pgrids, structure=vdata.structure, normalization=normalization
        )

    def reorient_axis(self) -> None:
        """Rorient the lattices.

        Change the orientation of the lattice vector so that: ``a`` points along the x-axis, ``b`` is in the xy-plane,
        ``c`` is in the positive-z halve of space

        """
        args: tuple[float, float, float, float, float, float] = (
            self.structure.lattice.abc + self.structure.lattice.angles
        )
        self.structure.lattice = Lattice.from_parameters(*args, vesta=True)

    def get_data_in_cube(self, s: float, ngrid: int, key: str = "total") -> npt.NDArray:
        """Return the charge density data sampled on a cube.

        Obtain a cubic basic cubic crop of the normalized charge density data.

        Parameters
        ----------
        s:
            The side length of the cube
        ngrid:
            Number of grid points in each direction
        key:
            The key to read from ``self.normalized_data``

        Returns
        -------
        NDArray:
            Regridded data in a ngrid x ngrid x ngrid array

        """
        grid_out = [ngrid, ngrid, ngrid]
        target_sc_lat_vecs = np.eye(3, 3) * s
        sc_mat = np.linalg.inv(self.structure.lattice.matrix) @ target_sc_lat_vecs
        _, res = get_sc_interp(self.normalized_data[key], sc_mat, grid_out)
        return res.reshape(grid_out)

    def get_transformed(
        self,
        sc_mat: npt.NDArray,
        grid_out: list[int] | int,
        origin: npt.ArrayLike = (0, 0, 0),
        up_sample: int = 1,
    ) -> "ChargeDensity":
        """Modify the structure and data and return a new object containing the reshaped data.

        Parameters
        ----------
        sc_mat:
            The transformation matrix to apply to the lattice vectors
        grid_out:
            The dimensions of the transformed grid
        origin:
            Origin of the new lattice in fractional coordinates of the input cell
        up_sample:
            The factor to scale up the sampling of the grid data using Fourier interpolation

        Returns
        -------
        ChargeDensity:
            The transformed ChargeDensity object

        """
        # warning if the sc_mat is not integer valued
        if not np.allclose(np.round(sc_mat), sc_mat):
            warnings.warn(
                "The `sc_mat` is not integer valued.\n"
                "Non-integer valued transformations are valid but will not product periodic structures, "
                "thus we cannot define a new Structure object.\n"
                "We will round the sc_mat to integer values for now but can implement functionality "
                "that returns a Molecule object in the future.",
            )
        sc_mat = np.round(sc_mat).astype(int)
        new_structure = self.structure.copy()
        new_structure.translate_sites(
            list(range(len(new_structure))), -np.array(origin)
        )
        new_structure = new_structure * sc_mat

        # determine the output grid
        lengths = new_structure.lattice.abc
        if isinstance(grid_out, int):
            ngrid = grid_out / new_structure.volume
            mult = (np.prod(lengths) / ngrid) ** (1 / 3)
            grid_out = [int(math.floor(max(l_ / mult, 1))) for l_ in lengths]

        pgrids = {}
        for k, pgrid in self.normalized_pgrids.items():
            new_pgrid = pgrid.get_transformed(
                sc_mat=sc_mat, grid_out=grid_out, origin=origin, up_sample=up_sample
            )
            pgrids[k] = _scaled_data(
                grid_data=new_pgrid,
                lattice=new_structure.lattice,
                normalization=self.normalization,
            )

        return ChargeDensity(
            pgrids=pgrids,
            structure=new_structure,
            normalization=self.normalization,
        )

    def to_Chgcar(self) -> Chgcar:
        """Convert the charge density to a ``pymatgen.io.vasp.outputs.Chgcar`` object.

        Scale and convert each key in the pgrids dictionary and create a ``Chgcar`` object

        Returns
        -------
        Chgcar:
            The charge density object

        """
        return self.to_VolumetricData(cls=Chgcar, normalization="vasp")

    def to_VolumetricData(
        self, cls=VolumetricData, normalization: str = "vasp"
    ) -> VolumetricData:
        """Convert the charge density to a ``pymatgen.io.vasp.outputs.VolumetricData`` object.

        Scale and convert each key in the pgrids dictionary and create a ``VolumetricData`` object

        Returns
        -------
        VolumetricData:
            The charge density object

        """
        struct = self.structure.copy()
        data_dict = {}
        for k, v in self.normalized_data.items():
            data_dict[k] = _scaled_data(
                v, lattice=self.structure.lattice, normalization=normalization
            )
        return cls(Poscar(structure=struct), data_dict)

    @classmethod
    def from_file(
        cls, filename: str, pmg_obj: VolumetricData = Chgcar
    ) -> "ChargeDensity":
        """Read a ChargeDensity object from a file.

        Parameters
        ----------
        filename:
            The filename of the ChargeDensity object
        pmg_obj:
            The pymatgen object to read from the file (default: Chgcar).
            the `from_file` method from this class will be called to read the file.

        Returns
        -------
            ChargeDensity: The ChargeDensity object

        """
        return cls.from_pmg(pmg_obj.from_file(filename))

    @classmethod
    def from_hdf5(
        cls, filename: str, pmg_obj: VolumetricData = Chgcar
    ) -> "ChargeDensity":
        """Read a ChargeDensity object from a hdf5 file.

        Parameters
        ----------
        filename:
            The filename of the ChargeDensity object
        pmg_obj:
            The pymatgen object to read from the file (default: Chgcar).
            the `from_file` method from this class will be called to read the file.

        Returns
        -------
            ChargeDensity: The ChargeDensity object

        """
        return cls.from_pmg(pmg_obj.from_hdf5(filename))


def get_matched_structure_mapping(
    uc_struct: Structure, sc_struct: Structure, sm: StructureMatcher | None = None
) -> tuple[npt.NDArray, npt.ArrayLike] | None:
    """Get the mapping of the supercell to the unit cell.

    Get the mapping from the supercell structure onto the base structure,
    Note: this only works for structures that are exactly matched.

    Parameters
    ----------
    uc_struct: host structure, smaller cell
    sc_struct: bigger cell
    sm: StructureMatcher instance

    Returns
    -------
    sc_m : supercell matrix to apply to s1 to get s2
    total_t : translation to apply on s1 * sc_m to get s2
    """
    if sm is None:
        sm = StructureMatcher(
            primitive_cell=False, comparator=ElementComparator(), attempt_supercell=True
        )
    s1, s2 = sm._process_species([sc_struct.copy(), uc_struct.copy()])
    trans = sm.get_transformation(s1, s2)
    if trans is None:
        return None
    sc, t, mapping = trans
    temp = s2.copy().make_supercell(sc)
    ii, jj = 0, mapping[0]
    vec = np.round(sc_struct[ii].frac_coords - temp[jj].frac_coords)
    return sc, t + vec


def get_volumetric_like_sc(
    vd: VolumetricData,
    sc_struct: Structure,
    grid_out: npt.ArrayLike,
    up_sample: int = 1,
    sm: StructureMatcher | None = None,
    normalization: str | None = "vasp",
):
    """Get the volumetric data in the supercell.

    Parameters
    ----------
    vd: VolumeData instance
    sc_struct: supercell structure.
    grid_out: grid size to output the volumetric data.
    up_sample: up sampling factor.
    sm: StructureMatcher instance
    normalization: normalization method for the volumetric data.
        default is "vasp" which assumes the normalization is the
        same as VASP's CHGCAR file. If None, no normalization is
        done.

    Returns
    -------
    VolumetricData: volumetric data in the supercell
    """
    trans = get_matched_structure_mapping(vd.structure, sc_struct=sc_struct, sm=sm)
    if trans is None:
        raise ValueError("Could not find a supercell mapping")
    sc_mat, total_t = trans
    cden = ChargeDensity.from_pmg(vd, normalization=normalization)
    orig = np.dot(total_t, sc_mat)
    cden_transformed = cden.get_transformed(
        sc_mat=sc_mat, origin=-orig, grid_out=grid_out, up_sample=up_sample
    )
    return cden_transformed.to_VolumetricData(
        cls=vd.__class__, normalization=normalization
    )


@deprecated
def multiply_aug(data_aug: list[str], factor: int) -> list[str]:
    """Update the data in the augmentation charge.

    The original idea here was to use to to speed up some vasp calculations for
    supercells by initializing the entire CHGCAR file.
    The current code does not deal with transformation of the Augementation charges after regridding.

    This is a naive way to multiply the Augmentation data in the CHGCAR,
    a real working implementation will require analysis of the PAW projection operators.
    However, even with such an implementation, the speed up will be minimal due to VASP's internal
    minimization algorithms.

    Parameters
    ----------
    data_aug:
        The original augmentation data from a CHGCAR
    factor:
        The multiplication factor (some integer number of times it gets repeated)

    Returns
    -------
    List[str]:
        Each line of the augmentation data.

    """
    res: list[str] = []
    cur_block: list[str] = []
    cnt = 0
    for ll in data_aug:
        if "augmentation" in ll:
            if cur_block:
                for _ in range(factor):
                    cnt += 1
                    cur_block[
                        0
                    ] = f"augmentation occupancies{cnt:>4}{cur_block[0].split()[-1]:>4}\n"
                    res.extend(cur_block)
            cur_block = [ll]
        else:
            cur_block.append(ll)
    for _ in range(factor):
        cnt += 1
        cur_block[
            0
        ] = f"augmentation occupancies{cnt:>4}{cur_block[0].split()[-1]:>4}\n"
        res.extend(cur_block)
    return res


def _normalize_data(
    grid_data: npt.NDArray, lattice: Lattice, normalization: str | None = "vasp"
) -> npt.NDArray:
    """Normalize the data to the number of electrons.

    Since different codes use different normalization methods for
    volumetric data we should convert them to the same units (electrons / Angstrom^3)

    Parameters
    ----------
    grid_data:
        The grid data to normalize
    lattice:
        The lattice that the grid data is represented on
    normalization:
        The normalization method defaults to vasp
        - None: no normalization
        - vasp:
            The standard charge density from VASP is given as (rho*V) such that:
                sum(rho)/NGRID = NELECT/vol
            so the real normalized rho should be:
                rho = (rho*vol)*NGRID/vol/vol
            where the second `/vol` account for the different number of electrons in
            different cells

    Returns
    -------
    NDArray:
        The normalized grid data

    """
    if normalization is None or normalization[0].lower() == "n":
        return grid_data
    elif normalization[0].lower() == "v":
        return grid_data / lattice.volume
    else:
        raise NotImplementedError("Not a valid normalization scheme")


def _scaled_data(
    grid_data: npt.NDArray, lattice: Lattice, normalization: str | None = "vasp"
) -> npt.NDArray:
    """Undo the normalization of the data.

    Parameters
    ----------
    grid_data:
        The grid data to unnormalize
    lattice:
        The lattice that the grid data is represented on
    normalization:
        The normalization method defaults to vasp

    Returns
    -------
    NDArray:
        The un-normalized grid data

    """
    if normalization is None or normalization[0].lower() == "n":
        return grid_data
    elif normalization[0].lower() == "v":
        return grid_data * lattice.volume
    else:
        raise NotImplementedError("Not a valid normalization scheme")



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

sys.path.insert(0, os.path.abspath("../.."))


# -- Project information -----------------------------------------------------

project = "mp-pyrho"
copyright = "2022, Materials Project Team"
author = "Jimmy-Xuan Shen"


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    "sphinx_autodoc_typehints",
    "m2r2",
    "nbsphinx",
    "nbsphinx_link",
    "autoapi.extension",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["Thumbs.db", ".DS_Store", "test*.py"]

# use type hints
autoapi_dirs = ["../../src"]
autoapi_add_toctree_entry = False
autoapi_python_class_content = "class"
# autoapi_options = ["members", "show-module-summary", "show-inheritance"]
autoapi_options = [
    "members",
    "private-members",
    "show-inheritance",
    "show-module-summary",
    "special-members",
    "imported-members",
]
# autodoc_typehints = "description"
# autoclass_content = 'init'

# better napoleon support
# napoleon_use_param = True
# napoleon_use_rtype = True
# napoleon_use_ivar = True
napoleon_numpy_docstring = True

# The suffix(es) of source filenames.
source_suffix = [".rst", ".md"]

# Ensure env.metadata[env.docname]['nbsphinx-link-target'] points relative to repo root:
nbsphinx_link_target_root = os.path.join(__file__, "..", "..")

nbsphinx_prolog = r"""
{% if env.metadata[env.docname]['nbsphinx-link-target'] %}
{% set docpath = env.metadata[env.docname]['nbsphinx-link-target'] %}
{% else %}
{% set docpath = env.doc2path(env.docname, base='docs/source/') %}
{% endif %}
.. only:: html
    .. role:: raw-html(raw)
        :format: html
    .. nbinfo::
        This page is available as a Jupyter notebook: `{{ docpath }}`__.
    __ https://github.com/materialsproject/jobflow/tree/main/{{ docpath }}"""


# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "furo"

# hide sphinx footer
html_show_sphinx = False
html_show_sourcelink = False

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
fonts = [
    "Lato",
    "-apple-system",
    "BlinkMacSystemFont",
    "Segoe UI",
    "Helvetica",
    "Arial",
    "sans-serif",
    "Apple Color Emoji",
    "Segoe UI Emoji",
]
html_static_path = ["_static"]
html_css_files = ["custom.css"]
html_theme_options = {
    "light_css_variables": {
        "admonition-font-size": "92%",
        "admonition-title-font-size": "92%",
        "font-stack": ",".join(fonts),
        "font-size--small": "92%",
        "font-size--small--2": "87.5%",
        "font-size--small--3": "87.5%",
        "font-size--small--4": "87.5%",
    },
    "dark_css_variables": {
        "admonition-font-size": "92%",
        "admonition-title-font-size": "92%",
        "font-stack": ",".join(fonts),
        "font-size--small": "92%",
        "font-size--small--2": "87.5%",
        "font-size--small--3": "87.5%",
        "font-size--small--4": "87.5%",
    },
}
html_title = "pyRho"



import numpy as np
import pytest
from pymatgen.io.vasp import Chgcar

from pyrho.charge_density import ChargeDensity, get_volumetric_like_sc


def test_charge_density(test_dir):
    cden_uc = ChargeDensity.from_file(test_dir / "CHGCAR.uc.vasp")
    cden_sc = ChargeDensity.from_file(test_dir / "CHGCAR.sc1.vasp")
    chgcar_sc = Chgcar.from_file(test_dir / "CHGCAR.sc1.vasp")
    cden_transformed = cden_uc.get_transformed(
        [[1, 1, 0], [1, -1, 0], [0, 0, 1]],
        grid_out=cden_sc.grid_shape,
        up_sample=2,
    )

    # check that the UC and SC normalized data looks the same
    assert cden_uc.normalized_data["total"].max() == pytest.approx(
        cden_sc.normalized_data["total"].max(),
        rel=1e-2,
    )
    # check that the transformed data looks the same as the SC data
    np.testing.assert_allclose(
        cden_transformed.normalized_data["total"],
        cden_sc.normalized_data["total"],
        atol=0.001,
        rtol=0.01,
    )

    chgcar_transformed = cden_transformed.to_Chgcar()

    np.testing.assert_allclose(
        chgcar_sc.data["total"], chgcar_transformed.data["total"], atol=0.1, rtol=0.01
    )  # since the chgcar is scaled, the absolute tolerance is 0.1

def test_mapping(test_dir):
    chgcar_uc = Chgcar.from_file(test_dir / "CHGCAR.uc.vasp")
    chgcar_sc = Chgcar.from_file(test_dir / "CHGCAR.sc1.vasp")
    grid_out = chgcar_sc.dim
    chgcar_out = get_volumetric_like_sc(chgcar_uc, chgcar_sc.structure, grid_out=grid_out, up_sample=2)
    np.testing.assert_allclose(chgcar_out.data["total"], chgcar_sc.data["total"], atol=0.1, rtol=0.01)